;*************************************************************************
; AMYCODERS BESTS STRIKE BACK
;-------------------------------------------------------------------------
; Coder Compo
;*************************************************************************
; Name:    Sum Pyramid calculator
; Type:    Riddle solver
; Mission: calculate the missing numbers of the sum pyramid
; Compos:               - shortest routine
;                       - fastest routine (on 060)
; Deadline:             20. June 2010, 20:20 h
; Send entries to:      scicco/scarab <scicco@scarab-amiga.com>
;*************************************************************************

	section	sp,code_f

sum_pyramid_compo:
	lea	template9(pc),a0		; <---- set your template here
	bsr	pyramid_prepare_riddle

	move.w	numrows(pc),d0
	lea	pyramid(pc),a0
	lea	your_space(pc),a1
	bsr	pyramid_riddle_solver

	bsr	pyramid_test_result
	rts

; this routine prepares a pyramid riddle
; in:	a0 = template to use
pyramid_prepare_riddle:
	* clear the pyramid
	lea	pyramid(pc),a1
	move.w	#32640-1,d0
.clearLoop
	move.l	#0,(a1)+
	dbf	d0,.clearLoop
	
	* init the pyramid with values
	lea	pyramid(pc),a1
	lea	numrows(pc),a2
	move.w	(a0)+,d1
	move.w	d1,(a2)		; copy num rows
	move.l	(a0)+,(a1)	; copy head value
	moveq	#0,d3
.findRow
	addq	#1,d3
	cmp.w	d3,d1
	beq.b	.rowFound
	lea	(a1,d3.w*4),a1
	bra.b	.findRow
.rowFound
	move.l	(a0)+,d2	; column
	beq.b	.templateDone
	subq	#1,d2
	lea	(a1,d2.w*4),a2
	move.l	(a0)+,(a2)	; copy value
	bra.b	.rowFound
.templateDone

	* copy the pyramid
	lea	pyramid(pc),a1
	lea	pyramid_copy(pc),a2
	move.w	#2080-1,d0
.copyLoop
	move.l	(a1)+,(a2)
	addq	#4,a2
	dbf	d0,.copyLoop
	rts

; this routine tests the pyramid against the copy
; and checks if the sums are correct
; fast code, not optimized, ignore ;)
pyramid_test_result:
	* first compare the copy with the original
	lea	pyramid(pc),a0
	lea	pyramid+4(pc),a2		; get row 2
	lea	pyramid_copy(pc),a1
	move.w	numrows(pc),d1
	subq.w	#1,d1
	moveq	#0,d2
.checkRow
	move.w	d2,d3
.checkCol
	move.l	(a0),d4
	beq.b	.errorHasUncalculatedValue
	move.l	(a1)+,d5
	beq.b	.noPresets
	cmp.l	d4,d5
	bne.b	.errorGivenVals
.noPresets
	tst.w	d1
	beq.b	.lastRow		; no sums at last row
	move.l	(a2)+,d6
	add.l	(a2),d6
	bcs.b	.errorHasOverflow
	cmp.l	d4,d6
	bne.b	.errorWrongSum
.lastRow
	addq	#4,a0
	dbf	d3,.checkCol
	addq	#4,a2
	addq.w	#1,d2
	dbf	d1,.checkRow
	moveq	#0,d0
	rts
.errorGivenVals:
	moveq	#1,d0
	rts
.errorWrongSum:
	moveq	#2,d0
	rts
.errorHasUncalculatedValue:
	moveq	#3,d0
	rts
.errorHasOverflow:
	moveq	#4,d0
	rts

;*************************************************************************
;*************************************************************************
;
; YOUR CODE GOES HERE
;
;*************************************************************************
;*************************************************************************

pyramid_riddle_solver:
	; D0 = Number of rows
	; A0 = Head
	; A1 = Zeroed mem

	move.l	d0,d1
	mulu.w	d1,d1
	sub.l	d0,d1
	lea.l	(a0,d1.l*2),a2	; Last row of pyramid
	lea.l	weights(pc),a6
	lea.l	(a6,d1.l*2),a6	; Weights for last row

	; Copy last row for reference
	move.l	d0,d2
	move.l	a2,a3
	move.l	a1,a4
.copyloop:
	move.l	(a3)+,(a4)+
	subq.l	#1,d2
	bne.b	.copyloop

.fill:
	; Put ones into vacant cells
	move.l	a4,a5	; Indices of vacant cells
	moveq.l	#0,d1	; Number of vacant cells
	moveq.l	#0,d2	; Index
	moveq.l	#0,d4	; Total head sum
.fillloop:
	move.l	(a1,d2.l*4),d3
	bne.b	.given
	moveq.l	#1,d3
	move.l	d3,(a2,d2.l*4)
	move.l	d2,(a5)+
	addq.l	#1,d1
.given:
	mulu.l	(a6,d2.l*4),d3	; Cell value * weight = contribution to head value
	add.l	d3,d4
	addq.l	#1,d2
	cmp.l	d0,d2
	blt.b	.fillloop

	tst.l	d1
	beq.w	.sumcorrected	; No vacant cells

	sub.l	(a0),d4	; Difference between head sum and actual head value (residue)
	beq.w	.sumcorrected	; Lucky strike match
	neg.l	d4	; Positive residue

	; Check for super-easy cases: first or last cell is vacant
	tst.l	(a1)
	bne.b	.firstnotvacant
	add.l	d4,(a2)
	bra.w	.sumcorrected
.firstnotvacant:
	tst.l	-4(a1,d0.l*4)
	bne.b	.lastnotvacant
	add.l	d4,-4(a2,d0.l*4)
	bra.w	.sumcorrected
.lastnotvacant:

	; Sort the vacant cells by decreasing weight.
	; Since the weights are increasing, then decreasing,
	; this can be done using a single merge
	move.l	a5,a3
	moveq.l	#0,d5
	move.l	(a4,d5.l*4),d2
.findlargest:
	move.l	(a6,d2.l*4),d3
	addq.l	#1,d5
	cmp.l	d1,d5
	bge.b	.largestfound
	move.l	(a4,d5.l*4),d2
	cmp.l	(a6,d2.l*4),d3
	blo.b	.findlargest
.largestfound:
	move.l	d5,d6
	subq.l	#1,d5	; Index of cell with largest weight

	move.l	(a4,d5.l*4),(a5)+
	subq.l	#1,d5
.merge:
	tst.l	d5
	bge.b	.leftvalid
	cmp.l	d1,d6
	bge.b	.mergedone
.right:	move.l	(a4,d6.l*4),(a5)+
	addq.l	#1,d6
	bra.b	.merge
.leftvalid:
	cmp.l	d1,d6
	blt.b	.mergecompare
.left:	move.l	(a4,d5.l*4),(a5)+
	subq.l	#1,d5
	bra.b	.merge
.mergecompare:
	move.l	(a4,d5.l*4),d2
	move.l	(a6,d2.l*4),d3
	move.l	(a4,d6.l*4),d2
	cmp.l	(a6,d2.l*4),d3
	bgt.b	.left
	blt.b	.right
	; Duplicate - take only one
	addq.l	#1,d6
	bra.b	.left
.mergedone:
	move.l	a3,a4
	move.l	a5,d1
	sub.l	a4,d1
	lsr.l	#2,d1

	; Eliminate all cells whose weight is larger than than the residue
	; or is a multiple of another vacant cell's weight

	move.l	d1,d7
.eliminate:
	subq.l	#1,d7
	ble.b	.elimdone
	move.l	(a4,d7.l*4),d2
	move.l	(a6,d2.l*4),d3
	move.l	a4,a3
	move.l	a4,a5
.elimloop:
	move.l	(a3)+,d2
	move.l	(a6,d2.l*4),d5
	move.l	d2,(a5)+
	cmp.l	d3,d5
	beq.b	.eliminate
	cmp.l	d4,d5
	bhi.b	.elim	; Larger than residue - remove
	move.l	d5,d6
	divu.l	d3,d5
	mulu.l	d3,d5
	cmp.l	d6,d5
	bne.b	.elimloop
.elim:
	; Eliminate
	subq.l	#4,a5
	subq.l	#1,d1
	subq.l	#1,d7
	bra.b	.elimloop
.elimdone:
	lea.l	(a4,d1.l*4),a5	; End of cell index list

	; Successively check if there exist one, two, three, etc.
	; vacant cell which can be modified to make the sum match

	moveq.l	#1,d7	; Number of correction cells
.ncellloop:
	move.l	a4,-(a7)
	bsr.b	.try
	move.l	(a7)+,a4
	beq.b	.foundmatch
	addq.l	#1,d7		; One more correction cell
	cmp.l	d1,d7
	ble.b	.ncellloop
	; Unsolvable
	illegal

.try:
	subq.l	#1,d7
	beq.b	.tryfindmatch
	; Iterate through possible values
.tryloop:
	move.l	(a4)+,d2	; Cell index
	move.l	(a6,d2.l*4),d3	; Weight
	sub.l	d3,d4		; New residue
	bcs.b	.iterationdone	; Cell value cannot be higher without overshooting the head value
	addq.l	#1,(a2,d2.l*4)	; Increase value
	move.l	a4,-(a7)
	bsr.b	.try		; Iterate on next vacant cell
	move.l	(a7)+,a4
	subq.l	#4,a4
	bne.b	.tryloop
	rts			; Match found - return through recursion

.iterationdone:
	; Restore cell value to 1 and proceed to next cell
	mulu.l	(a2,d2.l*4),d3
	add.l	d3,d4
	move.l	#1,(a2,d2.l*4)
	lea.l	(a4,d7.l*4),a3
	cmp.l	a5,a3
	blo.b	.tryloop	; Last cell which left space for the remaining number of correction cells
.endofiteration:
	addq.l	#1,d7
	rts			; Return to previous correction cell

.tryfindmatch:
	; Last correction cell - find a cell whose weight dvides the residue
	move.l	a4,a3
.findmatch:
	move.l	(a3)+,d2	; Cell index
	move.l	(a6,d2.l*4),d3	; Weight
	move.l	d4,d5
	divu.l	d3,d5
	move.l	d5,d6
	mulu.l	d3,d5
	cmp.l	d5,d4
	beq.b	.match
	cmp.l	a5,a3
	blo.b	.findmatch
	addq.l	#1,d7
.match:
	; Z set if we found a match
	rts

.foundmatch:
	; Add final correction value
	add.l	d6,(a2,d2.l*4)
.sumcorrected:

	; Last row now produces correct head value.
	; Calculate the rest of the pyramid.
	move.l	a2,a3
	move.l	d0,d3
.rowloop:
	subq.l	#1,d3
	move.l	d3,d2
	lea.l	(a3,d3.l*4),a5
.sumloop:
	move.l	(a5),d1
	add.l	-(a5),d1
	move.l	d1,-(a3)
	subq.l	#1,d2
	bne.b	.sumloop
	cmp.l	a0,a3
	bhi.b	.rowloop

	; We are done - phew!
	rts

weights:
	; Weights for each cell in pyramids up to size 31.
	; Each number indicates how much the head value would increase
	; if the corresponding cell value increased by 1.
	; This is simply Pascal's Triangle
	dc.l	1
	dc.l	1,1
	dc.l	1,2,1
	dc.l	1,3,3,1
	dc.l	1,4,6,4,1
	dc.l	1,5,10,10,5,1
	dc.l	1,6,15,20,15,6,1
	dc.l	1,7,21,35,35,21,7,1
	dc.l	1,8,28,56,70,56,28,8,1
	dc.l	1,9,36,84,126,126,84,36,9,1
	dc.l	1,10,45,120,210,252,210,120,45,10,1
	dc.l	1,11,55,165,330,462,462,330,165,55,11,1
	dc.l	1,12,66,220,495,792,924,792,495,220,66,12,1
	dc.l	1,13,78,286,715,1287,1716,1716,1287,715,286,78,13,1
	dc.l	1,14,91,364,1001,2002,3003,3432,3003,2002,1001,364,91,14,1
	dc.l	1,15,105,455,1365,3003,5005,6435,6435,5005,3003,1365,455,105,15,1
	dc.l	1,16,120,560,1820,4368,8008,11440,12870,11440,8008,4368,1820,560,120,16,1
	dc.l	1,17,136,680,2380,6188,12376,19448,24310,24310,19448,12376,6188,2380,680,136,17,1
	dc.l	1,18,153,816,3060,8568,18564,31824,43758,48620,43758,31824,18564,8568,3060,816,153,18,1
	dc.l	1,19,171,969,3876,11628,27132,50388,75582,92378,92378,75582,50388,27132,11628,3876,969,171,19,1
	dc.l	1,20,190,1140,4845,15504,38760,77520,125970,167960,184756,167960,125970,77520,38760,15504,4845,1140,190,20,1
	dc.l	1,21,210,1330,5985,20349,54264,116280,203490,293930,352716,352716,293930,203490,116280,54264,20349,5985,1330,210,21,1
	dc.l	1,22,231,1540,7315,26334,74613,170544,319770,497420,646646,705432,646646,497420,319770,170544,74613,26334,7315,1540,231,22,1
	dc.l	1,23,253,1771,8855,33649,100947,245157,490314,817190,1144066,1352078,1352078,1144066,817190,490314,245157,100947,33649,8855,1771,253,23,1
	dc.l	1,24,276,2024,10626,42504,134596,346104,735471,1307504,1961256,2496144,2704156,2496144,1961256,1307504,735471,346104,134596,42504,10626,2024,276,24,1
	dc.l	1,25,300,2300,12650,53130,177100,480700,1081575,2042975,3268760,4457400,5200300,5200300,4457400,3268760,2042975,1081575,480700,177100,53130,12650,2300,300,25,1
	dc.l	1,26,325,2600,14950,65780,230230,657800,1562275,3124550,5311735,7726160,9657700,10400600,9657700,7726160,5311735,3124550,1562275,657800,230230,65780,14950,2600,325,26,1
	dc.l	1,27,351,2925,17550,80730,296010,888030,2220075,4686825,8436285,13037895,17383860,20058300,20058300,17383860,13037895,8436285,4686825,2220075,888030,296010,80730,17550,2925,351,27,1
	dc.l	1,28,378,3276,20475,98280,376740,1184040,3108105,6906900,13123110,21474180,30421755,37442160,40116600,37442160,30421755,21474180,13123110,6906900,3108105,1184040,376740,98280,20475,3276,378,28,1
	dc.l	1,29,406,3654,23751,118755,475020,1560780,4292145,10015005,20030010,34597290,51895935,67863915,77558760,77558760,67863915,51895935,34597290,20030010,10015005,4292145,1560780,475020,118755,23751,3654,406,29,1
	dc.l	1,30,435,4060,27405,142506,593775,2035800,5852925,14307150,30045015,54627300,86493225,119759850,145422675,155117520,145422675,119759850,86493225,54627300,30045015,14307150,5852925,2035800,593775,142506,27405,4060,435,30,1

pyramid_riddle_solver_end:

	printv	pyramid_riddle_solver_end-pyramid_riddle_solver

;*************************************************************************
; END OF YOUR CODE

;*************************************************************************
; these are some preset templates. create your own for further
; tests.

template1:
	dc.w	5		; num rows
	dc.l	51		; head value
	dc.l	4,5		; value in last row: 4th column, value: 5
	dc.l	5,1		; value in last row: 5th column, value: 1
	dc.l	0

template2:
	dc.w	6		; num rows
	dc.l	95		; head value
	dc.l	2,3		; value in last row: 2nd column: value: 3
	dc.l	3,1		; value in last row: 3rd column: value: 1
	dc.l	0

template3:
	dc.w	20		; num rows
	dc.l	16584800	; head value
	dc.l	4,13		; value in last row: 4th column: value: 13
	dc.l	12,16		; value in last row: 12th column: value: 16
	dc.l	13,11		; value in last row: 13th column: value: 11
	dc.l	0

template4:
	dc.w	7		; num rows
	dc.l	283		; head value
	dc.l	1,4		; value in last row: 1st column: value: 4
	dc.l	2,3		; value in last row: 2nd column: value: 3
	dc.l	3,2		; value in last row: 3rd column: value: 2
	dc.l	5,2		; value in last row: 5th column: value: 2
	dc.l	6,3		; value in last row: 6th column: value: 3
	dc.l	7,3		; value in last row: 7th column: value: 3
	dc.l	0

template5:
	dc.w	30		; num rows
	dc.l	879981839	; head value
	dc.l	7,1
	dc.l	8,3
	dc.l	10,1
	dc.l	20,1
	dc.l	21,3
	dc.l	22,2
	dc.l	23,1
	dc.l	0

template6:
	dc.w	30		; num rows
	dc.l	879981839	; head value
	dc.l	1,87
	dc.l	7,1
	dc.l	8,3
	dc.l	10,1
	dc.l	20,1
	dc.l	21,3
	dc.l	22,2
	dc.l	23,1
	dc.l	0

template7:
	dc.w	5		; num rows
	dc.l	51		; head value
	dc.l	1,2		; value in last row: 4th column, value: 5
	dc.l	5,1		; value in last row: 5th column, value: 1
	dc.l	0

template8:
	dc.w	7		; num rows
	dc.l	105		; head value
	dc.l	1,1		; value in last row: 4th column, value: 5
	dc.l	7,1		; value in last row: 5th column, value: 1
	dc.l	0

	; Test case which stresses the algorithm
	; Three vacant cells need values larger than one.
template9:
	dc.w	29		; num rows
	dc.l	4242424242	; head value
	dc.l	1,4242
	dc.l	2,42
	dc.l	3,42
	dc.l	5,42
	dc.l	6,42
	dc.l	7,42
	dc.l	8,42
	dc.l	10,15
	dc.l	11,15
	dc.l	12,15
	dc.l	13,15
	dc.l	14,15
	dc.l	16,15
	dc.l	17,15
	dc.l	18,15
	dc.l	19,15
	dc.l	21,42
	dc.l	22,42
	dc.l	23,42
	dc.l	24,42
	dc.l	25,42
	dc.l	26,42
	dc.l	28,42
	dc.l	29,204
	dc.l	0

time_pyramid_0035:
	dc.w	28
	dc.l	160382576
	dc.l	1,2582569
	dc.l	5,45
	dc.l	6,126
	dc.l	12,1
	dc.l	14,1
	dc.l	15,1
	dc.l	18,1
	dc.l	22,16
	dc.l	23,78
	dc.l	28,926005

;*************************************************************************

numrows:
	dc.w	0

pyramid:
	dcb.l	2080

pyramid_copy:
	dcb.l	2080

your_space:
	dcb.b	1024*1024


; this copy is only made for testing purpose, ignore it

