;*************************************************************************
; AMYCODERS BESTS STRIKE BACK
;-------------------------------------------------------------------------
; Coder Compo
;*************************************************************************
; Name:    Sum Pyramid calculator
; Type:    Riddle solver
; Mission: calculate the missing numbers of the sum pyramid
; Compos:               - shortest routine
;                       - fastest routine (on 060)
; Deadline:             20. June 2010, 20:20 h
; Send entries to:      scicco/scarab <scicco@scarab-amiga.com>
;*************************************************************************

	section	sp,code_f

sum_pyramid_compo:
	lea	template5(pc),a0		; <---- set your template here
	bsr	pyramid_prepare_riddle

	move.w	numrows(pc),d0
	lea	pyramid(pc),a0
	lea	your_space(pc),a1
	bsr	pyramid_riddle_solver

	bsr	pyramid_test_result
	rts

; this routine prepares a pyramid riddle
; in:	a0 = template to use
pyramid_prepare_riddle:
	* clear the pyramid
	lea	pyramid(pc),a1
	move.w	#32640-1,d0
.clearLoop
	move.l	#0,(a1)+
	dbf	d0,.clearLoop
	
	* init the pyramid with values
	lea	pyramid(pc),a1
	lea	numrows(pc),a2
	move.w	(a0)+,d1
	move.w	d1,(a2)		; copy num rows
	move.l	(a0)+,(a1)	; copy head value
	moveq	#0,d3
.findRow
	addq	#1,d3
	cmp.w	d3,d1
	beq.b	.rowFound
	lea	(a1,d3.w*4),a1
	bra.b	.findRow
.rowFound
	move.l	(a0)+,d2	; column
	beq.b	.templateDone
	subq	#1,d2
	lea	(a1,d2.w*4),a2
	move.l	(a0)+,(a2)	; copy value
	bra.b	.rowFound
.templateDone

	* copy the pyramid
	lea	pyramid(pc),a1
	lea	pyramid_copy(pc),a2
	move.w	#2080-1,d0
.copyLoop
	move.l	(a1)+,(a2)
	addq	#4,a2
	dbf	d0,.copyLoop
	rts

; this routine tests the pyramid against the copy
; and checks if the sums are correct
; fast code, not optimized, ignore ;)
pyramid_test_result:
	* first compare the copy with the original
	lea	pyramid(pc),a0
	lea	pyramid+4(pc),a2		; get row 2
	lea	pyramid_copy(pc),a1
	move.w	numrows(pc),d1
	subq.w	#1,d1
	moveq	#0,d2
.checkRow
	move.w	d2,d3
.checkCol
	move.l	(a0),d4
	beq.b	.errorHasUncalculatedValue
	move.l	(a1)+,d5
	beq.b	.noPresets
	cmp.l	d4,d5
	bne.b	.errorGivenVals
.noPresets
	tst.w	d1
	beq.b	.lastRow		; no sums at last row
	move.l	(a2)+,d6
	add.l	(a2),d6
	cmp.l	d4,d6
	bne.b	.errorWrongSum
.lastRow
	addq	#4,a0
	dbf	d3,.checkCol
	addq	#4,a2
	addq.w	#1,d2
	dbf	d1,.checkRow
	moveq	#0,d0
	rts
.errorGivenVals:
	moveq	#1,d0
	rts
.errorWrongSum:
	moveq	#2,d0
	rts
.errorHasUncalculatedValue:
	moveq	#3,d0
	rts

;*************************************************************************
;*************************************************************************
;
; YOUR CODE GOES HERE
;
;*************************************************************************
;*************************************************************************

random	macro
	mulu.l	#54561,d7
	add.l	#16307,d7
	move.l	d7,\1
	swap.w	\1
	endm

pyramid_riddle_solver:
	; D0 = Number of rows
	; A0 = Head
	; A1 = Zeroed mem

	move.l	d0,d1
	mulu.w	d1,d1
	sub.l	d0,d1
	lea.l	(a0,d1.l*2),a2	; Last row of pyramid
	lea.l	weights(pc),a6
	lea.l	(a6,d1.l*2),a6	; Weights for last row

	; Copy last row for reference
	move.l	d0,d2
	move.l	a2,a3
	move.l	a1,a4
.copyloop:
	move.l	(a3)+,(a4)+
	subq.l	#1,d2
	bne.b	.copyloop

	move.l	#$ADAC0DE,d7	; Random seed
.fill:
	; Put random values into vacant cells
	move.l	a4,a5	; Indices of vacant cells
	moveq.l	#0,d1	; Number of vacant cells
	moveq.l	#0,d2	; Index
	moveq.l	#0,d4	; Total head sum
.fillloop:
	move.l	(a1,d2.l*4),d3
	bne.b	.given
.rval:	random	d3
	beq.b	.rval	; No zeros, please
	move.l	d3,(a2,d2.l*4)
	move.l	d2,(a5)+
	addq.l	#1,d1
.given:
	mulu.l	(a6,d2.l*4),d3	; Cell value * weight = contribution to head value
	add.l	d3,d4
	addq.l	#1,d2
	cmp.l	d0,d2
	blt.b	.fillloop

	tst.l	d1
	beq.b	.sumcorrected	; No vacant cells

	sub.l	(a0),d4	; Difference between head sum and actual head value (residue)
	beq.b	.sumcorrected	; Lucky strike match

	; Find lowest one bit in residue sum
	moveq.l	#0,d3
.residuelowestbit:
	btst.l	d3,d4
	bne.b	.residuelowestbitfound
	addq.l	#1,d3
	bra.b	.residuelowestbit
.residuelowestbitfound:
	; Calculate bit mask identifying suitable correction weights
	moveq.l	#2,d5
	lsl.l	d3,d5
	subq.l	#1,d5

	; Find a vacant cell suitable for correction (one where the modular
	; division between the residue and the weight is well-defined).
	; Choose at random and iterate until a suitable one is found.
	; If the puzzle is unsolvable, this will loop indefinitely.
.findcorrector:
	random	d2
	mulu.w	d1,d2
	swap.w	d2
	ext.l	d2
	move.l	(a4,d2.l*4),d2	; Index of vacant cell
	move.l	(a6,d2.l*4),d6	; Weight
	and.l	d5,d6
	beq.b	.findcorrector	; Unsuitable weight
	move.l	(a6,d2.l*4),d6	; Weight

	; Perform a modular division (modulo 2^32) between the residue
	; and the weight
	moveq.l	#0,d3
.weightlowestbit:
	btst.l	d3,d6
	bne.b	.weightlowestbitfound
	addq.l	#1,d3
	bra.b	.weightlowestbit
.weightlowestbitfound:

	moveq.l	#0,d1	; Accumulated result
	moveq.l	#1,d5	; Accumulator bit
.modulardivision:
	btst.l	d3,d4
	beq.b	.nextbit
	or.l	d5,d1
	sub.l	d6,d4
.nextbit:
	add.l	d5,d5
	add.l	d6,d6
	addq.l	#1,d3
	cmp.l	#32,d3
	blt.b	.modulardivision

	; Subtract correction value 
	sub.l	d1,(a2,d2.l*4)
	beq.w	.fill	; Corrected value is zero - try again
.sumcorrected:

	; Last row now produces correct head value.
	; Calculate the rest of the pyramid and check that no zeros arise.
	move.l	a2,a3
	move.l	d0,d3
.rowloop:
	subq.l	#1,d3
	move.l	d3,d2
	lea.l	(a3,d3.l*4),a5
.sumloop:
	move.l	(a5),d1
	add.l	-(a5),d1
	beq.w	.fill	; Intermediate value is zero - try again
	move.l	d1,-(a3)
	subq.l	#1,d2
	bne.b	.sumloop
	cmp.l	a0,a3
	bhi.b	.rowloop

	; We are done, and encountered no zeros - phew!
	rts

weights:
	; Weights (modulo 2^32) for each cell in pyramids up to size 64.
	; Each number indicates how much the head value would increase
	; if the corresponding cell value increased by 1.
	; This is simply Pascal's Triangle (modulo 2^32)
	dc.l	1
	dc.l	1,1
	dc.l	1,2,1
	dc.l	1,3,3,1
	dc.l	1,4,6,4,1
	dc.l	1,5,10,10,5,1
	dc.l	1,6,15,20,15,6,1
	dc.l	1,7,21,35,35,21,7,1
	dc.l	1,8,28,56,70,56,28,8,1
	dc.l	1,9,36,84,126,126,84,36,9,1
	dc.l	1,10,45,120,210,252,210,120,45,10,1
	dc.l	1,11,55,165,330,462,462,330,165,55,11,1
	dc.l	1,12,66,220,495,792,924,792,495,220,66,12,1
	dc.l	1,13,78,286,715,1287,1716,1716,1287,715,286,78,13,1
	dc.l	1,14,91,364,1001,2002,3003,3432,3003,2002,1001,364,91,14,1
	dc.l	1,15,105,455,1365,3003,5005,6435,6435,5005,3003,1365,455,105,15,1
	dc.l	1,16,120,560,1820,4368,8008,11440,12870,11440,8008,4368,1820,560,120,16,1
	dc.l	1,17,136,680,2380,6188,12376,19448,24310,24310,19448,12376,6188,2380,680,136,17,1
	dc.l	1,18,153,816,3060,8568,18564,31824,43758,48620,43758,31824,18564,8568,3060,816,153,18,1
	dc.l	1,19,171,969,3876,11628,27132,50388,75582,92378,92378,75582,50388,27132,11628,3876,969,171,19,1
	dc.l	1,20,190,1140,4845,15504,38760,77520,125970,167960,184756,167960,125970,77520,38760,15504,4845,1140,190,20,1
	dc.l	1,21,210,1330,5985,20349,54264,116280,203490,293930,352716,352716,293930,203490,116280,54264,20349,5985,1330,210,21,1
	dc.l	1,22,231,1540,7315,26334,74613,170544,319770,497420,646646,705432,646646,497420,319770,170544,74613,26334,7315,1540,231,22,1
	dc.l	1,23,253,1771,8855,33649,100947,245157,490314,817190,1144066,1352078,1352078,1144066,817190,490314,245157,100947,33649,8855,1771,253,23,1
	dc.l	1,24,276,2024,10626,42504,134596,346104,735471,1307504,1961256,2496144,2704156,2496144,1961256,1307504,735471,346104,134596,42504,10626,2024,276,24,1
	dc.l	1,25,300,2300,12650,53130,177100,480700,1081575,2042975,3268760,4457400,5200300,5200300,4457400,3268760,2042975,1081575,480700,177100,53130,12650,2300,300,25,1
	dc.l	1,26,325,2600,14950,65780,230230,657800,1562275,3124550,5311735,7726160,9657700,10400600,9657700,7726160,5311735,3124550,1562275,657800,230230,65780,14950,2600,325,26,1
	dc.l	1,27,351,2925,17550,80730,296010,888030,2220075,4686825,8436285,13037895,17383860,20058300,20058300,17383860,13037895,8436285,4686825,2220075,888030,296010,80730,17550,2925,351,27,1
	dc.l	1,28,378,3276,20475,98280,376740,1184040,3108105,6906900,13123110,21474180,30421755,37442160,40116600,37442160,30421755,21474180,13123110,6906900,3108105,1184040,376740,98280,20475,3276,378,28,1
	dc.l	1,29,406,3654,23751,118755,475020,1560780,4292145,10015005,20030010,34597290,51895935,67863915,77558760,77558760,67863915,51895935,34597290,20030010,10015005,4292145,1560780,475020,118755,23751,3654,406,29,1
	dc.l	1,30,435,4060,27405,142506,593775,2035800,5852925,14307150,30045015,54627300,86493225,119759850,145422675,155117520,145422675,119759850,86493225,54627300,30045015,14307150,5852925,2035800,593775,142506,27405,4060,435,30,1
	dc.l	1,31,465,4495,31465,169911,736281,2629575,7888725,20160075,44352165,84672315,141120525,206253075,265182525,300540195,300540195,265182525,206253075,141120525,84672315,44352165,20160075,7888725,2629575,736281,169911,31465,4495,465,31,1
	dc.l	1,32,496,4960,35960,201376,906192,3365856,10518300,28048800,64512240,129024480,225792840,347373600,471435600,565722720,601080390,565722720,471435600,347373600,225792840,129024480,64512240,28048800,10518300,3365856,906192,201376,35960,4960,496,32,1
	dc.l	1,33,528,5456,40920,237336,1107568,4272048,13884156,38567100,92561040,193536720,354817320,573166440,818809200,1037158320,1166803110,1166803110,1037158320,818809200,573166440,354817320,193536720,92561040,38567100,13884156,4272048,1107568,237336,40920,5456,528,33,1
	dc.l	1,34,561,5984,46376,278256,1344904,5379616,18156204,52451256,131128140,286097760,548354040,927983760,1391975640,1855967520,2203961430,2333606220,2203961430,1855967520,1391975640,927983760,548354040,286097760,131128140,52451256,18156204,5379616,1344904,278256,46376,5984,561,34,1
	dc.l	1,35,595,6545,52360,324632,1623160,6724520,23535820,70607460,183579396,417225900,834451800,1476337800,2319959400,3247943160,4059928950,242600354,242600354,4059928950,3247943160,2319959400,1476337800,834451800,417225900,183579396,70607460,23535820,6724520,1623160,324632,52360,6545,595,35,1
	dc.l	1,36,630,7140,58905,376992,1947792,8347680,30260340,94143280,254186856,600805296,1251677700,2310789600,3796297200,1272935264,3012904814,7562008,485200708,7562008,3012904814,1272935264,3796297200,2310789600,1251677700,600805296,254186856,94143280,30260340,8347680,1947792,376992,58905,7140,630,36,1
	dc.l	1,37,666,7770,66045,435897,2324784,10295472,38608020,124403620,348330136,854992152,1852482996,3562467300,1812119504,774265168,4285840078,3020466822,492762716,492762716,3020466822,4285840078,774265168,1812119504,3562467300,1852482996,854992152,348330136,124403620,38608020,10295472,2324784,435897,66045,7770,666,37,1
	dc.l	1,38,703,8436,73815,501942,2760681,12620256,48903492,163011640,472733756,1203322288,2707475148,1119983000,1079619508,2586384672,765137950,3011339604,3513229538,985525432,3513229538,3011339604,765137950,2586384672,1079619508,1119983000,2707475148,1203322288,472733756,163011640,48903492,12620256,2760681,501942,73815,8436,703,38,1
	dc.l	1,39,741,9139,82251,575757,3262623,15380937,61523748,211915132,635745396,1676056044,3910797436,3827458148,2199602508,3666004180,3351522622,3776477554,2229601846,203787674,203787674,2229601846,3776477554,3351522622,3666004180,2199602508,3827458148,3910797436,1676056044,635745396,211915132,61523748,15380937,3262623,575757,82251,9139,741,39,1
	dc.l	1,40,780,9880,91390,658008,3838380,18643560,76904685,273438880,847660528,2311801440,1291886184,3443288288,1732093360,1570639392,2722559506,2833032880,1711112104,2433389520,407575348,2433389520,1711112104,2833032880,2722559506,1570639392,1732093360,3443288288,1291886184,2311801440,847660528,273438880,76904685,18643560,3838380,658008,91390,9880,780,40,1
	dc.l	1,41,820,10660,101270,749398,4496388,22481940,95548245,350343565,1121099408,3159461968,3603687624,440207176,880414352,3302732752,4293198898,1260625090,249177688,4144501624,2840964868,2840964868,4144501624,249177688,1260625090,4293198898,3302732752,880414352,440207176,3603687624,3159461968,1121099408,350343565,95548245,22481940,4496388,749398,101270,10660,820,41,1
	dc.l	1,42,861,11480,111930,850668,5245786,26978328,118030185,445891810,1471442973,4280561376,2468182296,4043894800,1320621528,4183147104,3300964354,1258856692,1509802778,98712016,2690499196,1386962440,2690499196,98712016,1509802778,1258856692,3300964354,4183147104,1320621528,4043894800,2468182296,4280561376,1471442973,445891810,118030185,26978328,5245786,850668,111930,11480,861,42,1
	dc.l	1,43,903,12341,123410,962598,6096454,32224114,145008513,563921995,1917334783,1457037053,2453776376,2217109800,1069549032,1208801336,3189144162,264853750,2768659470,1608514794,2789211212,4077461636,4077461636,2789211212,1608514794,2768659470,264853750,3189144162,1208801336,1069549032,2217109800,2453776376,1457037053,1917334783,563921995,145008513,32224114,6096454,962598,123410,12341,903,43,1
	dc.l	1,44,946,13244,135751,1086008,7059052,38320568,177232627,708930508,2481256778,3374371836,3910813429,375918880,3286658832,2278350368,102978202,3453997912,3033513220,82206968,102758710,2571705552,3859955976,2571705552,102758710,82206968,3033513220,3453997912,102978202,2278350368,3286658832,375918880,3910813429,3374371836,2481256778,708930508,177232627,38320568,7059052,1086008,135751,13244,946,44,1
	dc.l	1,45,990,14190,148995,1221759,8145060,45379620,215553195,886163135,3190187286,1560661318,2990217969,4286732309,3662577712,1270041904,2381328570,3556976114,2192543836,3115720188,184965678,2674464262,2136694232,2136694232,2674464262,184965678,3115720188,2192543836,3556976114,2381328570,1270041904,3662577712,4286732309,2990217969,1560661318,3190187286,886163135,215553195,45379620,8145060,1221759,148995,14190,990,45,1
	dc.l	1,46,1035,15180,163185,1370754,9366819,53524680,260932815,1101716330,4076350421,455881308,255911991,2981982982,3654342725,637652320,3651370474,1643337388,1454552654,1013296728,3300685866,2859429940,516191198,4273388464,516191198,2859429940,3300685866,1013296728,1454552654,1643337388,3651370474,637652320,3654342725,2981982982,255911991,455881308,4076350421,1101716330,260932815,53524680,9366819,1370754,163185,15180,1035,46,1
	dc.l	1,47,1081,16215,178365,1533939,10737573,62891499,314457495,1362649145,883099455,237264433,711793299,3237894973,2341358411,4291995045,4289022794,999740566,3097890042,2467849382,19015298,1865148510,3375621138,494612366,494612366,3375621138,1865148510,19015298,2467849382,3097890042,999740566,4289022794,4291995045,2341358411,3237894973,711793299,237264433,883099455,1362649145,314457495,62891499,10737573,1533939,178365,16215,1081,47,1
	dc.l	1,48,1128,17296,194580,1712304,12271512,73629072,377348994,1677106640,2245748600,1120363888,949057732,3949688272,1284286088,2338386160,4286050543,993796064,4097630608,1270772128,2486864680,1884163808,945802352,3870233504,989224732,3870233504,945802352,1884163808,2486864680,1270772128,4097630608,993796064,4286050543,2338386160,1284286088,3949688272,949057732,1120363888,2245748600,1677106640,377348994,73629072,12271512,1712304,194580,17296,1128,48,1
	dc.l	1,49,1176,18424,211876,1906884,13983816,85900584,450978066,2054455634,3922855240,3366112488,2069421620,603778708,939007064,3622672248,2329469407,984879311,796459376,1073435440,3757636808,76061192,2829966160,521068560,564490940,564490940,521068560,2829966160,76061192,3757636808,1073435440,796459376,984879311,2329469407,3622672248,939007064,603778708,2069421620,3366112488,3922855240,2054455634,450978066,85900584,13983816,1906884,211876,18424,1176,49,1
	dc.l	1,50,1225,19600,230300,2118760,15890700,99884400,536878650,2505433700,1682343578,2994000432,1140566812,2673200328,1542785772,266712016,1657174359,3314348718,1781338687,1869894816,536104952,3833698000,2906027352,3351034720,1085559500,1128981880,1085559500,3351034720,2906027352,3833698000,536104952,1869894816,1781338687,3314348718,1657174359,266712016,1542785772,2673200328,1140566812,2994000432,1682343578,2505433700,536878650,99884400,15890700,2118760,230300,19600,1225,50,1
	dc.l	1,51,1275,20825,249900,2349060,18009460,115775100,636763050,3042312350,4187777278,381376714,4134567244,3813767140,4215986100,1809497788,1923886375,676555781,800720109,3651233503,2405999768,74835656,2444758056,1962094776,141626924,2214541380,2214541380,141626924,1962094776,2444758056,74835656,2405999768,3651233503,800720109,676555781,1923886375,1809497788,4215986100,3813767140,4134567244,381376714,4187777278,3042312350,636763050,115775100,18009460,2349060,249900,20825,1275,51,1
	dc.l	1,52,1326,22100,270725,2598960,20358520,133784560,752538150,3679075400,2935122332,274186696,220976662,3653367088,3734785944,1730516592,3733384163,2600442156,1477275890,156986316,1762265975,2480835424,2519593712,111885536,2103721700,2356168304,134115464,2356168304,2103721700,111885536,2519593712,2480835424,1762265975,156986316,1477275890,2600442156,3733384163,1730516592,3734785944,3653367088,220976662,274186696,2935122332,3679075400,752538150,133784560,20358520,2598960,270725,22100,1326,52,1
	dc.l	1,53,1378,23426,292825,2869685,22957480,154143080,886322710,136646254,2319230436,3209309028,495163358,3874343750,3093185736,1170335240,1168933459,2038859023,4077718046,1634262206,1919252291,4243101399,705461840,2631479248,2215607236,164922708,2490283768,2490283768,164922708,2215607236,2631479248,705461840,4243101399,1919252291,1634262206,4077718046,2038859023,1168933459,1170335240,3093185736,3874343750,495163358,3209309028,2319230436,136646254,886322710,154143080,22957480,2869685,292825,23426,1378,53,1
	dc.l	1,54,1431,24804,316251,3162510,25827165,177100560,1040465790,1022968964,2455876690,1233572168,3704472386,74539812,2672562190,4263520976,2339268699,3207792482,1821609773,1417012956,3553514497,1867386394,653595943,3336941088,552119188,2380529944,2655206476,685600240,2655206476,2380529944,552119188,3336941088,653595943,1867386394,3553514497,1417012956,1821609773,3207792482,2339268699,4263520976,2672562190,74539812,3704472386,1233572168,2455876690,1022968964,1040465790,177100560,25827165,3162510,316251,24804,1431,54,1
	dc.l	1,55,1485,26235,341055,3478761,28989675,202927725,1217566350,2063434754,3478845654,3689448858,643077258,3779012198,2747102002,2641115870,2307822379,1252093885,734434959,3238622729,675560157,1125933595,2520982337,3990537031,3889060276,2932649132,740769124,3340806716,3340806716,740769124,2932649132,3889060276,3990537031,2520982337,1125933595,675560157,3238622729,734434959,1252093885,2307822379,2641115870,2747102002,3779012198,643077258,3689448858,3478845654,2063434754,1217566350,202927725,28989675,3478761,341055,26235,1485,55,1
	dc.l	1,56,1540,27720,367290,3819816,32468436,231917400,1420494075,3281001104,1247313112,2873327216,37558820,127122160,2231146904,1093250576,653970953,3559916264,1986528844,3973057688,3914182886,1801493752,3646915932,2216552072,3584630011,2526742112,3673418256,4081575840,2386646136,4081575840,3673418256,2526742112,3584630011,2216552072,3646915932,1801493752,3914182886,3973057688,1986528844,3559916264,653970953,1093250576,2231146904,127122160,37558820,2873327216,1247313112,3281001104,1420494075,231917400,32468436,3819816,367290,27720,1540,56,1
	dc.l	1,57,1596,29260,395010,4187106,36288252,264385836,1652411475,406527883,233346920,4120640328,2910886036,164680980,2358269064,3324397480,1747221529,4213887217,1251477812,1664619236,3592273278,1420709342,1153442388,1568500708,1506214787,1816404827,1905193072,3460026800,2173254680,2173254680,3460026800,1905193072,1816404827,1506214787,1568500708,1153442388,1420709342,3592273278,1664619236,1251477812,4213887217,1747221529,3324397480,2358269064,164680980,2910886036,4120640328,233346920,406527883,1652411475,264385836,36288252,4187106,395010,29260,1596,57,1
	dc.l	1,58,1653,30856,424270,4582116,40475358,300674088,1916797311,2058939358,639874803,59019952,2736559068,3075567016,2522950044,1387699248,776651713,1666141450,1170397733,2916097048,961925218,718015324,2574151730,2721943096,3074715495,3322619614,3721597899,1070252576,1338314184,51542064,1338314184,1070252576,3721597899,3322619614,3074715495,2721943096,2574151730,718015324,961925218,2916097048,1170397733,1666141450,776651713,1387699248,2522950044,3075567016,2736559068,59019952,639874803,2058939358,1916797311,300674088,40475358,4582116,424270,30856,1653,58,1
	dc.l	1,59,1711,32509,455126,5006386,45057474,341149446,2217471399,3975736669,2698814161,698894755,2795579020,1517158788,1303549764,3910649292,2164350961,2442793163,2836539183,4086494781,3878022266,1679940542,3292167054,1001127530,1501691295,2102367813,2749250217,496883179,2408566760,1389856248,1389856248,2408566760,496883179,2749250217,2102367813,1501691295,1001127530,3292167054,1679940542,3878022266,4086494781,2836539183,2442793163,2164350961,3910649292,1303549764,1517158788,2795579020,698894755,2698814161,3975736669,2217471399,341149446,45057474,5006386,455126,32509,1711,59,1
	dc.l	1,60,1770,34220,487635,5461512,50063860,386206920,2558620845,1898240772,2379583534,3397708916,3494473775,17770512,2820708552,919231760,1780032957,312176828,984365050,2628066668,3669549751,1262995512,677140300,4293294584,2502818825,3604059108,556650734,3246133396,2905449939,3798423008,2779712496,3798423008,2905449939,3246133396,556650734,3604059108,2502818825,4293294584,677140300,1262995512,3669549751,2628066668,984365050,312176828,1780032957,919231760,2820708552,17770512,3494473775,3397708916,2379583534,1898240772,2558620845,386206920,50063860,5461512,487635,34220,1770,60,1
	dc.l	1,61,1830,35990,521855,5949147,55525372,436270780,2944827765,161894321,4277824306,1482325154,2597215395,3512244287,2838479064,3739940312,2699264717,2092209785,1296541878,3612431718,2002649123,637577967,1940135812,675467588,2501146113,1811910637,4160709842,3802784130,1856616039,2408905651,2283168208,2283168208,2408905651,1856616039,3802784130,4160709842,1811910637,2501146113,675467588,1940135812,637577967,2002649123,3612431718,1296541878,2092209785,2699264717,3739940312,2838479064,3512244287,2597215395,1482325154,4277824306,161894321,2944827765,436270780,55525372,5949147,521855,35990,1830,61,1
	dc.l	1,62,1891,37820,557845,6471002,61474519,491796152,3381098545,3106722086,144751331,1465182164,4079540549,1814492386,2055756055,2283452080,2144237733,496507206,3388751663,614006300,1320113545,2640227090,2577713779,2615603400,3176613701,18089454,1677653183,3668526676,1364432873,4265521690,397106563,271369120,397106563,4265521690,1364432873,3668526676,1677653183,18089454,3176613701,2615603400,2577713779,2640227090,1320113545,614006300,3388751663,496507206,2144237733,2283452080,2055756055,1814492386,4079540549,1465182164,144751331,3106722086,3381098545,491796152,61474519,6471002,557845,37820,1891,62,1
	dc.l	1,63,1953,39711,595665,7028847,67945521,553270671,3872894697,2192853335,3251473417,1609933495,1249755417,1599065639,3870248441,44240839,132722517,2640744939,3885258869,4002757963,1934119845,3960340635,922973573,898349883,1497249805,3194703155,1695742637,1051212563,737992253,1334987267,367660957,668475683,668475683,367660957,1334987267,737992253,1051212563,1695742637,3194703155,1497249805,898349883,922973573,3960340635,1934119845,4002757963,3885258869,2640744939,132722517,44240839,3870248441,1599065639,1249755417,1609933495,3251473417,2192853335,3872894697,553270671,67945521,7028847,595665,39711,1953,63,1

pyramid_riddle_solver_end:

	printv	pyramid_riddle_solver_end-pyramid_riddle_solver

;*************************************************************************
; END OF YOUR CODE

;*************************************************************************
; these are some preset templates. create your own for further
; tests.

template1:
	dc.w	5		; num rows
	dc.l	51		; head value
	dc.l	4,5		; value in last row: 4th column, value: 5
	dc.l	5,1		; value in last row: 5th column, value: 1
	dc.l	0

template2:
	dc.w	6		; num rows
	dc.l	95		; head value
	dc.l	2,3		; value in last row: 2nd column: value: 3
	dc.l	3,1		; value in last row: 3rd column: value: 1
	dc.l	0

template3:
	dc.w	20		; num rows
	dc.l	16584800	; head value
	dc.l	4,13		; value in last row: 4th column: value: 13
	dc.l	12,16		; value in last row: 12th column: value: 16
	dc.l	13,11		; value in last row: 13th column: value: 11
	dc.l	0

template4:
	dc.w	7		; num rows
	dc.l	283		; head value
	dc.l	1,4		; value in last row: 1st column: value: 4
	dc.l	2,3		; value in last row: 2nd column: value: 3
	dc.l	3,2		; value in last row: 3rd column: value: 2
	dc.l	5,2		; value in last row: 5th column: value: 2
	dc.l	6,3		; value in last row: 6th column: value: 3
	dc.l	7,3		; value in last row: 7th column: value: 3
	dc.l	0

template5:
	dc.w	64		; num rows
	dc.l	879981839	; head value
	dc.l	7,1
	dc.l	8,3
	dc.l	10,1
	dc.l	20,1
	dc.l	30,3
	dc.l	33,2
	dc.l	34,1
	dc.l	0


;*************************************************************************

numrows:
	dc.w	0

pyramid:
	dcb.l	2080

pyramid_copy:
	dcb.l	2080

your_space:
	dcb.b	1024*1024


; this copy is only made for testing purpose, ignore it

