;*************************************************************************
; AMYCODERS BESTS STRIKE BACK
;-------------------------------------------------------------------------
; Coder Compo
;*************************************************************************
; Name:    Sum Pyramid calculator
; Type:    Riddle solver
; Mission: calculate the missing numbers of the sum pyramid
; Compos:               - shortest routine
;                       - fastest routine (on 060)
; Deadline:             20. June 2010, 20:20 h
; Send entries to:      scicco/scarab <scicco@scarab-amiga.com>
;*************************************************************************

	section	sp,code_f

sum_pyramid_compo:
	lea	template4(pc),a0		; <---- set your template here
	bsr	pyramid_prepare_riddle

	move.w	numrows(pc),d0
	lea	pyramid(pc),a0
	lea	your_space(pc),a1
	bsr	pyramid_riddle_solver

	bsr	pyramid_test_result
	rts

; this routine prepares a pyramid riddle
; in:	a0 = template to use
pyramid_prepare_riddle:
	* clear the pyramid
	lea	pyramid(pc),a1
	move.w	#32640-1,d0
.clearLoop
	move.l	#0,(a1)+
	dbf	d0,.clearLoop
	
	* init the pyramid with values
	lea	pyramid(pc),a1
	lea	numrows(pc),a2
	move.w	(a0)+,d1
	move.w	d1,(a2)		; copy num rows
	move.l	(a0)+,(a1)	; copy head value
	moveq	#0,d3
.findRow
	addq	#1,d3
	cmp.w	d3,d1
	beq.b	.rowFound
	lea	(a1,d3.w*4),a1
	bra.b	.findRow
.rowFound
	move.l	(a0)+,d2	; column
	beq.b	.templateDone
	subq	#1,d2
	lea	(a1,d2.w*4),a2
	move.l	(a0)+,(a2)	; copy value
	bra.b	.rowFound
.templateDone

	* copy the pyramid
	lea	pyramid(pc),a1
	lea	pyramid_copy(pc),a2
	move.w	#2080-1,d0
.copyLoop
	move.l	(a1)+,(a2)
	addq	#4,a2
	dbf	d0,.copyLoop
	rts

; this routine tests the pyramid against the copy
; and checks if the sums are correct
; fast code, not optimized, ignore ;)
pyramid_test_result:
	* first compare the copy with the original
	lea	pyramid(pc),a0
	lea	pyramid+4(pc),a2		; get row 2
	lea	pyramid_copy(pc),a1
	move.w	numrows(pc),d1
	subq.w	#1,d1
	moveq	#0,d2
.checkRow
	move.w	d2,d3
.checkCol
	move.l	(a0),d4
	beq.b	.errorHasUncalculatedValue
	move.l	(a1)+,d5
	beq.b	.noPresets
	cmp.l	d4,d5
	bne.b	.errorGivenVals
.noPresets
	tst.w	d1
	beq.b	.lastRow		; no sums at last row
	move.l	(a2)+,d6
	add.l	(a2),d6
	cmp.l	d4,d6
	bne.b	.errorWrongSum
.lastRow
	addq	#4,a0
	dbf	d3,.checkCol
	addq	#4,a2
	addq.w	#1,d2
	dbf	d1,.checkRow
	moveq	#0,d0
	rts
.errorGivenVals:
	moveq	#1,d0
	rts
.errorWrongSum:
	moveq	#2,d0
	rts
.errorHasUncalculatedValue:
	moveq	#3,d0
	rts

;*************************************************************************
;*************************************************************************
;
; YOUR CODE GOES HERE
;
;*************************************************************************
;*************************************************************************

pyramid_riddle_solver:
	; D0 = Number of rows
	; A0 = Head
	; A1 = Zeroed mem

	move.l	(a0),d7		; Save head value
	move.l	d0,d1
	mulu.w	d1,d1
	sub.l	d0,d1
	lea.l	(a0,d1.l*2),a4	; Last row
	lea.l	(a4,d0.l*4),a2	; End of pyramid
.recurse:
	cmp.l	a4,a2
	beq.b	.sum		; Recursed through all vacant cells
	tst.l	-(a2)
	bne.b	.recurse
.tryloop:
	; Try next value for this cell
	addq.l	#1,(a2)
	beq.b	.ret		; Wrap-around: No values of this cell give a valid solution
				; for the current values of the cells to the right
	move.l	a2,-(a7)
	bsr.b	.recurse	; Recursicely try out all values for the next cell to the left
	move.l	(a7)+,a2
	bcs.b	.tryloop	; There is no "C set or Z clear" branch,
	bne.b	.tryloop	; So we need two branches

	; We are done, but it doesn't hurt checking again...

.sum:
	; Sum up values in the pyramid
	move.l	a4,a3
	move.l	d0,d3
.rowloop:
	subq.l	#1,d3
	move.l	d3,d2
	lea.l	(a3,d3.l*4),a5
.sumloop:
	move.l	(a5),d1
	add.l	-(a5),d1
	beq.b	.ret		; Zeros are not allowed (note: C flag will also be set)
	move.l	d1,-(a3)
	subq.l	#1,d2
	bne.b	.sumloop
	cmp.l	a0,a3
	bhi.b	.rowloop
	cmp.l	d7,d1		; Correct head value?
.ret:
	; Return condition: Z set and C clear if done
	rts
pyramid_riddle_solver_end:

	printv	pyramid_riddle_solver_end-pyramid_riddle_solver

;*************************************************************************
; END OF YOUR CODE

;*************************************************************************
; these are some preset templates. create your own for further
; tests.

template1:
	dc.w	5		; num rows
	dc.l	51		; head value
	dc.l	4,5		; value in last row: 4th column, value: 5
	dc.l	5,1		; value in last row: 5th column, value: 1
	dc.l	0

template2:
	dc.w	6		; num rows
	dc.l	95		; head value
	dc.l	2,3		; value in last row: 2nd column: value: 3
	dc.l	3,1		; value in last row: 3rd column: value: 1
	dc.l	0

template3:
	dc.w	20		; num rows
	dc.l	16584800	; head value
	dc.l	4,13		; value in last row: 4th column: value: 13
	dc.l	12,16		; value in last row: 12th column: value: 16
	dc.l	13,11		; value in last row: 13th column: value: 11
	dc.l	0

template4:
	dc.w	7		; num rows
	dc.l	283		; head value
	dc.l	1,4		; value in last row: 1st column: value: 4
	dc.l	2,3		; value in last row: 2nd column: value: 3
	dc.l	3,2		; value in last row: 3rd column: value: 2
	dc.l	5,2		; value in last row: 5th column: value: 2
	dc.l	6,3		; value in last row: 6th column: value: 3
	dc.l	7,3		; value in last row: 7th column: value: 3
	dc.l	0

template5:
	dc.w	64		; num rows
	dc.l	879981839	; head value
	dc.l	7,1
	dc.l	8,3
	dc.l	10,1
	dc.l	20,1
	dc.l	30,3
	dc.l	33,2
	dc.l	34,1
	dc.l	0


;*************************************************************************

numrows:
	dc.w	0

pyramid:
	dcb.l	2080

pyramid_copy:
	dcb.l	2080

your_space:
	dcb.b	1024*1024


; this copy is only made for testing purpose, ignore it

