/*
 *
 * Pyramid calculation base on Genetic Programming
 * Qick & Dirty
 *
 * Coded for the Coders Cometition 2010 initialed by Scicco/Scarab
 *
 * (C) Enrico Ehrich alias Sniper/S 2010
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define ANIMAL_COUNT 64
#define CROSSOVER_CHANCE 3
#define MUTATION_CHANCE 10
#define JITTER_CHANCE 3

static int orig_pyramid[65535];
static int pyramid[65535];
static int dnabuffer[65535];


int template1[]=
{
     5,
     51,
     4, 5,
     5, 1,
     0
};

int template2[]=
{
     6,
     95,
     2, 3,
     3, 1,
     0
};

int template3[]=
{
     20,
     16584800,
     4, 13,
     12, 16,
     13, 11,
     0
};

int template4[]=
{
     7,
     283,
     1, 4,
     2, 3,
     3, 2,
     5, 2,
     6, 3,
     7, 3,
     0
};

int template5[]=
{
     64,
     879981839,
     7, 1,
     8, 3,
     10, 1,
     20, 1,
     //30, 3,
     //33, 2,
     //34, 1,
     0
};


int randrange( int n )
{
    unsigned long r = (rand() + (rand()<<4)) & RAND_MAX;

    if( n == 0 )
        return 0;

    r = r * (n-1) * 256;
    r = r / RAND_MAX + 128;
    r = r / 256;

    return r;
}

static int *pyramid_getLastRow( int *p, int rows );


void pyramid_init( int *template )
{
    int rows = template[0];
    int head = template[1];
    int *pairs = &template[2];
    int *last_row = pyramid_getLastRow( orig_pyramid + 1, rows );

    memset( orig_pyramid, 0, 65535 );
    orig_pyramid[0] = rows;
    orig_pyramid[1] = head;


    for(;;)
    {
        int col = *pairs++;
        if( col == 0)
            break;

        col--;

        if( col < rows )
            last_row[col] = *pairs++;
        else
            pairs++;
    }
}

void pyramid_copy( int *p_out, const int *p_in, int rows )
{
    int i, j;

    for( j=0; j<rows+1; j++ )
    {
        for( i=0; i<j; i++ )
        {
            *p_out++ = *p_in++;
        }
    }
}

static int *pyramid_getLastRow( int *p, int rows )
{
    int i;

    // goto last row
    for( i=0; i<rows; i++ )
        p += i;

    return p;
}


int pyramid_calc( int *p, int rows )
{
    int i,j;

    p = pyramid_getLastRow( p, rows );

    for( j=rows; j>2; j-- )
    {
        int *p_out = p-j+1;

        for( i=0; i<j-1; i++ )
        {
            p_out[i] = p[i] + p[i+1];
        }

        p = p_out;
    }

    return p[0]+ p[1];
}

int pyramid_solvable( int *p, int rows )
{
    int i;
    int *p1 = pyramid_getLastRow( p, rows );

    /*
    int minval = 1<<rows;

    // no solution
    if( minval <  p[0] )
        return 0;
    */

    for( i=0; i<rows; i++ )
    {
        if( p1[i] == 0 )
            p1[i] = 1;
    }

    return (pyramid_calc( p, rows ) <= p[0]);
}


void pyramid_print( int *p, int rows )
{
    int i, j;
    for( j=1; j<=rows; j++ )
    {
        for( i=0; i<j; i++)
        {
            printf("%12d", *p++);
        }
        printf("\n");
    }
}


void dna_jitter( int *p, int *po, int columns, int m )
{
    int i;
    int maxval = m/columns;

    for( i=0; i<columns; i++ )
    {
        if( po[i] == 0 )
        {
            // jitter
            int r = randrange( JITTER_CHANCE );
            if( (r == 0) && (p[i] < maxval) )
                p[i]++;
            else if( ( r == 1 ) && ( p[i]>1 ) )
                p[i]--;
#if 1
            // mutation
            if( randrange(MUTATION_CHANCE) == 0 )
            {
                p[i] = randrange(maxval) + 1;
            }
#endif
        }
    }
}

void dna_init( int *buf,  int *orig_pyramid )
{
    int i, j;
    int rows = orig_pyramid[0];
    int m = orig_pyramid[1]/rows; // for speed / rows
    int *last_row_orig =  pyramid_getLastRow( orig_pyramid + 1, rows );

    for( j=0; j<ANIMAL_COUNT; j++ )
    {
        for( i=0; i<rows; i++ )
        {
            if( last_row_orig[i] == 0 )
                buf[i] = randrange(m) + 1; // can also filled with 1
            else
            	buf[i] = last_row_orig[i];

        }
        buf += rows;
    }
}


int dna_step( int *dnabuf, int *pyramid, int *orig_pyramid )
{
	static int count = 0;
	static int best = 0x7FFFFFFF;

	double avg = 0;
    int *buf = dnabuf;
    int i, a, current, bestidx = 0, tmp;
    int rows = orig_pyramid[0];
    int target = orig_pyramid[1];
    int *last_row = pyramid_getLastRow( pyramid, rows );
    int *last_row_orig = pyramid_getLastRow( orig_pyramid+1, rows );

    int newbest = 0;

    for( a=0; a<ANIMAL_COUNT; a++ )
    {
        memcpy( last_row,  buf, rows*sizeof(int) );
        tmp = pyramid_calc( pyramid, rows );
        current = target - tmp;
        if( current < 0 )
            current = -current;

        avg += current;

        // best found?
        if( current < best )
        {
            best = current;
            bestidx = a;
            newbest = 1;

            if( best == 0 )
                break;
        }
        // jitter
        else
        {
            dna_jitter( buf, last_row_orig, rows, target  );
        }

        // next animAL
        buf += rows;
    }

    // copy best
    memcpy( last_row,  dnabuf + (bestidx*rows), rows*sizeof(int) );

    // crossover
    buf = dnabuf;
    for( a=0; a<ANIMAL_COUNT; a++ )
    {
    	// do not overwrite the best
    	if( a != bestidx )
    	{

    		int r1 = randrange(ANIMAL_COUNT);

    		for( i=0; i<rows; i++ )
    		{
    			if( 0 == randrange(CROSSOVER_CHANCE) )
    			{
    				buf[(r1*rows)+i] = last_row[i];
    			}
    		}
    	}
    }

    count++;

    if( newbest )
    {
    	printf( "avg=%f best=%d\n", avg/ANIMAL_COUNT, best );
    	fflush( stdout );
    }

    return best;
}



void pyramid_solutionDNA()
{
    int rows = orig_pyramid[0];
    int target = orig_pyramid[1];

    printf( "start rows=%d, target=%d\n", rows, target );

    pyramid_copy( pyramid, orig_pyramid + 1, rows );

    if( pyramid_solvable( pyramid, rows ) )
    {
        // init animals
        dna_init( dnabuffer, orig_pyramid );

        while( 1 )
        {
            int best = dna_step( dnabuffer, pyramid, orig_pyramid );
            if( 0 == best )
            {
                if( pyramid_calc( pyramid, rows ) == target )
                {
                    printf( "\nsolution:\n" );
                    pyramid_print( pyramid, rows );
                    fflush( stdout );
#if 0
                    dna_init( dnabuffer, orig_pyramid );
#else
                    break;
#endif
                }
                else
                {
                    printf( "someting was wrong calculated!!!!\n" );
                    break;
                }

            }
            else
            {
            }
        }
    }
    else
    {
        printf( "not resoveable\n" );
        pyramid_print( pyramid, rows );
    }
}


// do not works with complicated pyramids
void pyramid_solutionBruteforce()
{
    int i, j;
    int rows = orig_pyramid[0];
    int target = orig_pyramid[1];
    int *last_row = pyramid_getLastRow( pyramid, rows );
    int *orig_last_row = pyramid_getLastRow( orig_pyramid+1, rows );

    printf( "start rows=%d, target=%d\n", rows, target );

    pyramid_copy( pyramid, orig_pyramid + 1, rows );
    for( i=0; i<rows; i++ )
    {
        if( last_row[i]<=0)
            last_row[i] = 1;
    }

    pyramid_copy( dnabuffer, pyramid, rows );

    if( pyramid_solvable( pyramid, rows ) )
    {
        int best = 0x7FFFFFF;
        while(1)
        {
            pyramid_copy( pyramid, dnabuffer, rows  );

            // for each row
            for( i=0; i<rows; i++ )
            {
                if( orig_last_row[i] == 0 )
                {
                    for( j=0; j<1; j++)
                    {
                        int found = 0;
                        do
                        {
                            if( j == 0 )
                            {
                                last_row[i]++;
                            }
                            else
                            {
                                last_row[i]--;
                            }

                            if( last_row[i]<=0)
                                last_row[i] = 1;

                            int current = pyramid_calc( pyramid, rows );
                            current -= target;
                            if( current < 0 )
                                current = -current;

                            if( current < best )
                            {
                                pyramid_copy( dnabuffer, pyramid, rows );
                                best = current;
                                found = 1;

                                if( best == 0)
                                {
                                    printf( "\nsolution:\n" );
                                    pyramid_print( pyramid, rows );
                                    fflush( stdout );
                                    return;
                                }
                            }
                            else
                            {
                                // restore it
                                pyramid_copy( pyramid, dnabuffer, rows  );
                                found = 0;
                            }
                        } while( found );
                    }
                }
            }
        }
    }
    else
        printf( "not resoveable\n" );
}



int main( int argc, char **argv )
{
    pyramid_init( template1 );

    pyramid_solutionDNA();
    //pyramid_solutionBruteforce();

    printf( "ende\n" );
    return 0;
}

/* ============== HISTORY ============== */
/* $Log: $
 *
 */
